@@ -4,9 +4,13 @@ Base script class for spiders crawl managers.
 import abc
 import json
 import logging
+from argparse import Namespace
+from typing import Optional, List, Tuple, Dict, NewType, cast, Generator
 
 from bloom_filter import BloomFilter
+from typing_extensions import TypedDict, NotRequired
 
+from shub_workflow.script import JobKey, JobDict
 from shub_workflow.base import WorkFlowManager
 from shub_workflow.utils import hashstr
 
@@ -15,23 +19,36 @@ _LOG = logging.getLogger(__name__)
 _LOG.setLevel(logging.INFO)
 
 
-def get_spider_args_from_params(params):
-    spider_args = params.copy()
-    for key in "spider", "units", "project_id", "tags":
+SpiderArgs = NewType("SpiderArgs", Dict[str, str])
+
+
+class JobParams(TypedDict):
+    units: NotRequired[int]
+    tags: NotRequired[List[str]]
+    job_settings: NotRequired[Dict[str, str]]
+    spider_args: Optional[Dict[str, str]]
+    project_id: Optional[int]
+
+
+class FullJobParams(JobParams):
+    spider: NotRequired[Optional[str]]
+
+
+def get_spider_args_from_params(params: JobParams) -> SpiderArgs:
+    spider_args = cast(SpiderArgs, params.copy())
+    for key in tuple(JobParams.__annotations__):
         spider_args.pop(key, None)
     return spider_args
 
 
-def get_jobseq(tags):
-    jobseq, repetition = 0, 0
+def get_jobseq(tags: List[str]) -> Tuple[int, int]:
     for tag in tags:
         if tag.startswith("JOBSEQ="):
             tag = tag.replace("JOBSEQ=", "")
             jobseq, *rep = tag.split(".r")
-            rep.append(0)
-            repetition = int(rep[0])
-            jobseq = int(jobseq)
-    return jobseq, repetition
+            rep.append("0")
+            return int(jobseq), int(rep[0])
+    return 0, 0
 
 
 class CrawlManager(WorkFlowManager):
@@ -40,15 +57,15 @@ class CrawlManager(WorkFlowManager):
     finished. Close reason of the manager will be inherited from spider one.
     """
 
-    spider = None
+    spider: Optional[str] = None
 
     def __init__(self):
         super().__init__()
         self.__close_reason = None
 
         # running jobs represents the state of a crawl manager.
-        # a dict job key : (spider, spider_args_override)
-        self._running_job_keys = {}
+        # a dict job key : (spider, job_args_override)
+        self._running_job_keys: Dict[JobKey, Tuple[str, JobParams]] = {}
 
     @property
     def description(self):
@@ -62,60 +79,77 @@ class CrawlManager(WorkFlowManager):
         self.argparser.add_argument("--job-settings", help="Job settings dict in json format", default="{}")
         self.argparser.add_argument("--units", help="Set default number of ScrapyCloud units for each job", type=int)
 
-    def parse_args(self):
+    def parse_args(self) -> Namespace:
         args = super().parse_args()
         if self.spider is None:
             self.spider = args.spider
         return args
 
-    def get_spider_args(self, override=None):
+    def get_spider_args(self, override: JobParams) -> JobParams:
         spider_args = json.loads(self.args.spider_args)
-        if override:
-            spider_args.update(override)
+        spider_args.update(override)
         return spider_args
 
-    def get_job_settings(self, override=None):
+    def get_job_settings(self, override: Optional[Dict[str, str]] = None) -> Dict[str, str]:
         job_settings = json.loads(self.args.job_settings)
         if override:
             job_settings.update(override)
         return job_settings
 
-    def schedule_spider(self, spider=None, spider_args_override=None):
-        if spider_args_override is not None:
-            spider_args_override.copy()
+    def schedule_spider_with_jobargs(
+        self,
+        job_args_override: JobParams,
+        spider: Optional[str] = None,
+    ) -> Optional[JobKey]:
         spider = spider or self.spider
-        spider_args = self.get_spider_args(spider_args_override)
-        job_settings_override = spider_args.pop("job_settings", None)
-        job_settings = self.get_job_settings(job_settings_override)
-        spider_args["job_settings"] = job_settings
-        spider_args["units"] = spider_args.get("units", self.args.units)
-        jobkey = super().schedule_spider(spider, **spider_args)
-        if jobkey is not None:
-            self._running_job_keys[jobkey] = spider, spider_args_override
-        return jobkey
+        if spider is not None:
+            spider_args = self.get_spider_args(job_args_override)
+            job_settings_override = spider_args.get("job_settings", None)
+            spider_args["job_settings"] = self.get_job_settings(job_settings_override)
+            spider_args["units"] = spider_args.get("units", self.args.units)
+            jobkey = super().schedule_spider(spider, **spider_args)
+            if jobkey is not None:
+                self._running_job_keys[jobkey] = spider, job_args_override
+            return jobkey
+        return None
 
-    def check_running_jobs(self):
+    def schedule_spider(
+        self,
+        spider: Optional[str] = None,
+        tags: Optional[List[str]] = None,
+        units: Optional[int] = None,
+        project_id: Optional[int] = None,
+        **kwargs,
+    ) -> Optional[JobKey]:
+        job_args_override = kwargs.pop("job_args_override", None)
+        if spider is None or job_args_override is not None:
+            job_args_override = job_args_override or {}
+            job_args_override.update(kwargs)
+            return self.schedule_spider_with_jobargs(job_args_override, spider)
+        return super().schedule_spider(spider, **kwargs)
+
+    def check_running_jobs(self) -> Dict[JobKey, str]:
         outcomes = {}
         running_job_keys = list(self._running_job_keys)
         while running_job_keys:
             jobkey = running_job_keys.pop()
             if (outcome := self.is_finished(jobkey)) is not None:
-                spider, spider_args_override = self._running_job_keys.pop(jobkey)
+                spider, job_args_override = self._running_job_keys.pop(jobkey)
                 if outcome in self.failed_outcomes:
                     _LOG.warning(f"Job {jobkey} finished with outcome {outcome}.")
-                    if spider_args_override is not None:
-                        spider_args_override = spider_args_override.copy()
-                    self.bad_outcome_hook(spider, outcome, spider_args_override, jobkey)
+                    if job_args_override is not None:
+                        job_args_override = job_args_override.copy()
+                    self.bad_outcome_hook(spider, outcome, job_args_override, jobkey)
                 outcomes[jobkey] = outcome
         _LOG.info(f"There are {len(self._running_job_keys)} jobs still running.")
 
         return outcomes
 
-    def bad_outcome_hook(self, spider, outcome, spider_args_override, jobkey):
+    def bad_outcome_hook(self, spider: str, outcome: str, job_args_override: JobParams, jobkey: JobKey):
         if self.__close_reason is None:
             self.__close_reason = outcome
 
-    def workflow_loop(self):
+    def workflow_loop(self) -> bool:
         outcomes = self.check_running_jobs()
         if outcomes:
             return False
@@ -123,19 +157,12 @@ class CrawlManager(WorkFlowManager):
             self.schedule_spider()
         return True
 
-    def resume_workflow(self):
-        running_jobs = []
-        count = 0
-        for job in self.get_owned_jobs(state=["running", "pending"], meta=["spider_args", "tags", "spider"]):
-            key = job["key"]
-            spider_args_override = job.get("spider_args", {}).copy()
-            spider_args_override["tags"] = job["tags"]
-            self._running_job_keys[key] = job["spider"], spider_args_override
-            _LOG.info(f"added running job {key}")
-            running_jobs.append(job)
-            count += 1
-        _LOG.info(f"Added a total of {count} running jobs.")
-        return running_jobs
+    def resume_running_job_hook(self, job: JobDict):
+        key = job["key"]
+        job_args_override = cast(JobParams, job.get("spider_args", {}).copy())
+        job_args_override["tags"] = job["tags"]
+        self._running_job_keys[key] = job["spider"], job_args_override
+        _LOG.info(f"added running job {key}")
 
     def on_close(self):
         job = self.get_job()
@@ -150,10 +177,10 @@ class PeriodicCrawlManager(CrawlManager):
     parameters. Don't forget to set loop mode.
     """
 
-    def bad_outcome_hook(self, spider, outcome, spider_args_override, jobkey):
+    def bad_outcome_hook(self, spider: str, outcome: str, job_args_override: JobParams, jobkey: JobKey):
         pass
 
-    def workflow_loop(self):
+    def workflow_loop(self) -> bool:
         self.check_running_jobs()
         if not self._running_job_keys:
             self.schedule_spider()
@@ -178,24 +205,23 @@ class GeneratorCrawlManager(CrawlManager):
     def __init__(self):
         super().__init__()
         self.__parameters_gen = self.set_parameters_gen()
-        self.__additional_jobs = []
+        self.__additional_jobs: List[FullJobParams] = []
         self.__next_job_seq = 1
         self.__jobids = BloomFilter(max_elements=self.MAX_TOTAL_JOBS, error_rate=0.001)
 
-    def bad_outcome_hook(self, spider, outcome, spider_args_override, jobkey):
+    def bad_outcome_hook(self, spider: str, outcome: str, job_args_override: JobParams, jobkey: JobKey):
         pass
 
-    def add_job(self, spider, spider_args_override):
-        params = (spider_args_override or {}).copy()
+    def add_job(self, spider: str, job_args_override: JobParams):
+        params = cast(FullJobParams, (job_args_override or {}).copy())
         params["spider"] = spider
         self.__additional_jobs.append(params)
 
     @abc.abstractmethod
-    def set_parameters_gen(self):
-        for i in []:
-            yield i
+    def set_parameters_gen(self) -> Generator[FullJobParams, None, None]:
+        ...
 
-    def __add_jobseq_tag(self, params):
+    def __add_jobseq_tag(self, params: FullJobParams):
         tags = params.setdefault("tags", [])
         jobseq_tag = None
         for tag in tags:
@@ -211,7 +237,7 @@ class GeneratorCrawlManager(CrawlManager):
             jobseq_tag = f"JOBSEQ={jobseq:010d}.r{repn + 1}"
         tags.append(jobseq_tag)
 
-    def workflow_loop(self):
+    def workflow_loop(self) -> bool:
         self.check_running_jobs()
         while len(self._running_job_keys) < self.max_running_jobs:
             try:
@@ -225,7 +251,7 @@ class GeneratorCrawlManager(CrawlManager):
                 jobid = self.get_job_id({"spider": spider, "spider_args": spider_args})
                 if jobid not in self.__jobids:
                     self.__add_jobseq_tag(next_params)
-                    self.schedule_spider(spider=spider, spider_args_override=next_params)
+                    self.schedule_spider(spider=spider, job_args_override=next_params)
                     self.__jobids.add(jobid)
             except StopIteration:
                 if self._running_job_keys:
@@ -234,7 +260,7 @@ class GeneratorCrawlManager(CrawlManager):
         return True
 
     @staticmethod
-    def get_job_id(job):
+    def get_job_id(job: JobDict) -> str:
         jdict = job.get("spider_args", {}).copy()
         jdict["spider"] = job["spider"]
         for k, v in jdict.items():
@@ -242,21 +268,20 @@ class GeneratorCrawlManager(CrawlManager):
         jid = json.dumps(jdict, sort_keys=True)
         return hashstr(jid)
 
-    def resume_workflow(self):
-        for job in super().resume_workflow():
-            jobid = self.get_job_id(job)
+    def resume_running_job_hook(self, job: JobDict):
+        super().resume_running_job_hook(job)
+        jobid = self.get_job_id(job)
+        self.__jobids.add(jobid)
+        self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)
+
+    def resume_finished_job_hook(self, job: JobDict):
+        jobid = self.get_job_id(job)
+        if jobid not in self.__jobids:
             self.__jobids.add(jobid)
-            self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)
-        count = 0
-        diff_count = 0
-        for job in self.get_owned_jobs(spider=self.spider, state=["finished"], meta=["spider_args", "tags", "spider"]):
-            jobid = self.get_job_id(job)
-            if jobid not in self.__jobids:
-                diff_count += 1
-                self.__jobids.add(jobid)
-            self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)
-            count += 1
-        _LOG.info(f"Added a total of {count} completed jobs ({diff_count} different ones)")
+        self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)
+
+    def resume_workflow(self):
+        super().resume_workflow()
         _LOG.info(f"Next job sequence number: {self.__next_job_seq}")
 
     def on_close(self):
