@@ -134,15 +134,17 @@ class CrawlManagerTest(TestCase):
         self.assertFalse(mocked_super_schedule_spider.called)
 
     @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
-    def test_schedule_spider_with_resume_dd(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
+    def test_schedule_spider_with_resume(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
         with script_args(["myspider", "--flow-id=3a20"]):
             manager = TestManager()
 
         mocked_get_jobs_side_effect = [
             # the resumed job
             [{"tags": ["FLOW_ID=3a20"], "key": "999/10/1"}],
-            # the owned job
+            # the owned running job
             [{"spider": "myspider", "key": "999/1/1", "tags": ["FLOW_ID=3a20", "PARENT_NAME=test"]}],
+            # the owned finished jobs
+            [],
         ]
         mocked_get_jobs.side_effect = mocked_get_jobs_side_effect
         manager._WorkFlowManager__on_start()
@@ -191,13 +193,13 @@ class CrawlManagerTest(TestCase):
             [{"tags": ["FLOW_ID=3a20", "NAME=test"], "key": "999/10/1"}],
             # the not owned job
             [{"key": "999/1/1", "tags": ["FLOW_ID=3a20", "PARENT_NAME=testa"]}],
+            # owned finished jobs
+            [],
         ]
         manager._WorkFlowManager__on_start()
         self.assertTrue(manager.is_resumed)
         self.assertEqual(len(manager._running_job_keys), 0)
 
-        self.assertEqual(manager.get_jobs.call_count, 2)
-
         # first loop: no spider, schedule one.
         manager.is_finished = lambda x: None
         mocked_super_schedule_spider.side_effect = ["999/1/2"]
@@ -400,6 +402,8 @@ class CrawlManagerTest(TestCase):
             project_id=999,
             tags=["CHECKED", "JOBSEQ=0000000001"],
         )
+        jobid = GeneratorCrawlManager.get_job_id({"spider": "myspider", "spider_args": {"argA": "valA"}})
+        self.assertTrue(jobid in manager._GeneratorCrawlManager__jobids)
 
     @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
     def test_schedule_spider_list_with_resume(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
